# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys

sys.path.insert(0, os.path.abspath('../ptovnetlab/'))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'ptovnetlab'
copyright = '2025, Mencken Davidson'
author = 'Mencken Davidson'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.viewcode',  # Add a link to the Python source code for classes, functions etc.
    'autoapi.extension',
    'myst_parser',
    'sphinx.ext.autodoc',  # Core Sphinx library for auto html doc generation from docstrings
    'sphinx.ext.autodoc.typehints',
    'sphinx.ext.autosummary',  # Create neat summary tables for modules/classes/methods etc
    'sphinx.ext.intersphinx',  # Link to other project's documentation (see mapping below)
    'sphinx.ext.napoleon',
    'sphinx_autodoc_typehints', # Automatically document param types (less noise in class signature)
    ]

source_suffix = [".rst",  ".md"]
templates_path = ['_templates']
exclude_patterns = ['_build', '_templates', 'Thumbs.db', '.DS_Store']
html_show_sourcelink = False  # Remove 'view source code' from top of page (for html, not python)
html_theme = 'sphinx_rtd_theme'
autodoc_typehints = "description"
autoapi_template_dir = "_templates/autoapi"
#autodoc_class_signature = "separated"
autoapi_own_page_level = "module"
autoapi_dirs = ['../ptovnetlab/']
autoapi_type = "python"
autoapi_options = [
    "members",
    "undoc-members",
    "show-inheritance",
    "show-module-summary",
    "imported-members",
]

html_css_files = [
    "css/custom.css",
]

#on_rtd = os.environ.get("READTHEDOCS", None) == "True"
#if not on_rtd:  # only import and set the theme if we're building docs locally
#    import sphinx_rtd_theme
#    html_theme = "sphinx_rtd_theme"
#    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
#html_css_files = ["readthedocs-custom.css"] # Override some CSS settings

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_static_path = ['_static']

napoleon_google_docstring = True
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = True
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = True
napoleon_use_admonition_for_notes = True
napoleon_use_admonition_for_references = True
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True
napoleon_preprocess_types = False
napoleon_type_aliases = None
napoleon_attr_annotations = True

def contains(seq, item):
    return item in seq

def prepare_jinja_env(jinja_env) -> None:
    jinja_env.tests["contains"] = contains

autoapi_prepare_jinja_env = prepare_jinja_env

html_theme_options = {
    # Toc options
    'collapse_navigation': True,
    'sticky_navigation': True,
    'navigation_depth': -1,
    'includehidden': True,
    'titles_only': False
}



"""eosPoller.py

Uses asyncio.to_thread and gather functions to poll multiple switches concurrently """

import asyncio
from concurrent.futures import ThreadPoolExecutor
import pyeapi


def invoker(switchlist_in: list, uname_in: str, passwd_in: str,
            runtype_in: str) -> tuple[list, list, list]:
    """Run synchronously; provide entry-point for module; manage the asyncio eventloop;
    invoke async/threaded functions to do the real work.

    Parameters
    ---------
    switchlist_in : list
        List of Arista switches to be interrogated
    uname_in : str
        The username credential for authenticating to the switch
    passwd_in : str
        The password credential for authenticating to the switch

    Returns
    -------
    svsout : list
        Values of the switch that we're interested in
    lldpsout : list
        List of connections inferred from LLDP tables on the switches
    cfgsout : list
        List-of-lists holding all of the switch's configurations
            'all_configs_out[0]' is the configuration of the first item in switchlist
            'all_configs_out[0][0] is the 1st line of the 1st switch's configuration
    """

    svsout, lldpsout, cfgsout = asyncio.run(main(switchlist_in, uname_in, passwd_in,
                                            runtype_in))
    return svsout, lldpsout, cfgsout


async def main(switchlist_in2: list, uname_in2: str, passwd_in2: str,
               runtype_in2: str) -> tuple[list, list, list]:
    """Connect to  switches and get the data that we want from them.  Use asyncio
    to_thread function to enable concurrent processing of multiple switches.

    Parameters
    ---------
    switchlist_in2 : list
        List of Arista switches to be interrogated
    uname_in2 : str
        The username credential for authenticating to the switch
    passw_in2 : str
        The password credential for authenticating to the switch

    Returns
    -------
    sw_vals_out2 : list
        Values of the switch that we're interested in
    lldpngbrs_out2 : list
        List of connections inferred from LLDP tables on the switches
    all_configs_out2 : list
        List-of-lists holding all of the switch's configurations
            'all_configs_out2[0]' is the configuration of the first item in switchlist
            'all_configs_out2[0][0] is the 1st line of the 1st switch's configuration
    """
    # Set the maximum number of worker-threads we're willing to use
    loop = asyncio.get_running_loop()
    loop.set_default_executor(ThreadPoolExecutor(max_workers=20))
    # Initialize the lists we will eventually return
    switchcount2 = len(switchlist_in2)
    sw_vals_out2 = [[]] * (switchcount2)
    allconfigs_out2 = [[]] * (switchcount2)
    lldpngbrs_out2 = []
    tasks = []

    print('Polling Arista switches via EOS API..')
    # Enumerate the switchlist, spawning a new asyncio thread/task to get data from each
    for sw_cntr2, value2 in enumerate(switchlist_in2):
        coro = asyncio.to_thread(get_sw_data, value2, uname_in2, passwd_in2, sw_cntr2)
        tasks.append(asyncio.create_task(coro))

    # Gather the data from all the EAPI polling threads

    answers = [] * (switchcount2)
    answers = await asyncio.gather(*tasks)
    print('Finished polling switches.')
    print('')
    for val in answers:
        # Item [0] in each "answer" is the list of switch properties (switch_vals)
        sw_vals_out2[val[3]] = val[0]
        # Item [0] in each "answer" is the list of LLDP neighbors for the switch
        for i, entry in enumerate(val[1]):
            lldpngbrs_out2.append(entry)
        allconfigs_out2[val[3]] = val[2]
    return sw_vals_out2, lldpngbrs_out2, allconfigs_out2


async def get_sw_data(switch3: str, uname_in3: str, passwd_in3: str, sw_cntr3_in: int) -> tuple[Switch, list, list]:
    """Connect to a switch and get the data that we want from it."""
    # Clear any existing pyeapi.client.config
    pyeapi.client.config.clear()
    # Build the pyeapi.client.config object required for connecting to the switch
    pyeapi.client.config.add_connection(switch3, host=switch3, transport='https',
                                        username=uname_in3, password=passwd_in3)
    # Connect to the switch
    node = pyeapi.connect_to(switch3)
    # Get JSON-formatted results of several 'show...' commands
    eos_output = node.enable(("show version", "show lldp neighbors",
                             "show lldp local-info"), format="json")
    # Extract data from the "show" commands' output
    eos_output_model = eos_output[0]["result"]["modelName"]
    eos_output_ver = eos_output[0]["result"]["version"]
    eos_output_mac = eos_output[0]["result"]["systemMacAddress"]
    eos_output_serial = eos_output[0]["result"]["serialNumber"]
    eos_output_lldpname = eos_output[2]["result"]["systemName"]

    # Create a Switch object
    switch = Switch(
        name=switch3,
        model=eos_output_model,
        eos_version=eos_output_ver,
        system_mac=eos_output_mac,
        serial_number=eos_output_serial,
        lldp_system_name=eos_output_lldpname,
        ethernet_interfaces=0,  # Will be updated later
        gns3_template_id='',  # Will be updated later
        gns3_node_id='',  # Will be updated later
        docker_container_id='',  # Will be updated later
    )

    # Create this_sw_lldpnbrs list to return
    this_sw_lldpnbrs = []
    for value in eos_output[1]["result"]["lldpNeighbors"]:
        this_sw_lldpnbrs.append([str(eos_output_lldpname), str(value["port"]),
                                 str(value["neighborDevice"]),
                                 str(value["neighborPort"])])

    # Create this_sw_cfg list to return
    this_sw_cfg = node.running_config.splitlines()
    print("Finished polling: " + switch3)
    return switch, this_sw_lldpnbrs, this_sw_cfg


#!/usr/bin/env python3
"""
GNS3 Network Topology Management Module

This module provides asynchronous functionality for creating and 
configuring network topologies in GNS3, including device creation, 
configuration, and network infrastructure setup.
"""

import asyncio
import logging
from io import BytesIO
import tarfile
from typing import List, Dict, Optional, Union, Tuple, Any

import aiohttp
import aiodocker
from aiodocker import Docker as docker

from ptovnetlab.data_classes import Switch, Connection

# Configure logging
logging.basicConfig(
    level=logging.INFO, 
    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s'
)
logger = logging.getLogger(__name__)

class GNS3WorkerError(Exception):
    """Custom exception for GNS3 Worker related errors."""
    pass

class ContainerConfigurationError(GNS3WorkerError):
    """Exception raised for errors during container configuration."""
    pass

def invoker(
    servername: str, 
    gns3_url: str, 
    switches: List[Switch],
    prj_id: str, 
    connections: List[Connection]
) -> str:
    """
    Synchronous entry point for creating GNS3 project nodes and connections.

    This function provides a synchronous wrapper for the asynchronous main_job,
    allowing the module to be called from synchronous code.

    Args:
        servername (str): The name of the aiohttp.ClientSession object
        gns3_url (str): The URL for the GNS3 server
        switches (List[Switch]): List of Switch objects to be emulated
        prj_id (str): The GNS3 project ID
        connections (List[Connection]): List of connections to make between nodes

    Returns:
        str: Status message indicating completion or error

    Raises:
        GNS3WorkerError: If there are issues during the GNS3 project setup
    """
    try:
        logger.info('Initiating GNS3 project node and connection creation')
        result = asyncio.run(
            main_job(servername, gns3_url, switches, prj_id, connections)
        )
        logger.info('GNS3 project setup completed successfully')
        return result
    except Exception as e:
        logger.error(f'GNS3 project setup failed: {e}')
        raise GNS3WorkerError(f'Project setup failed: {e}')

async def main_job(
    servername: str, 
    gns3_url: str, 
    switches: List[Switch],
    prj_id: str, 
    connections: List[Connection]
) -> str:
    """
    Asynchronously create GNS3 nodes, configure containers, and establish connections.

    Args:
        servername (str): The name of the aiohttp.ClientSession object
        gns3_url (str): The URL for the GNS3 server
        switches (List[Switch]): List of Switch objects to be emulated
        prj_id (str): The GNS3 project ID
        connections (List[Connection]): List of connections to make between nodes

    Returns:
        str: Status message indicating completion
    """
    logger.info('Creating nodes in the GNS3 project.')
    
    # Configure session timeout
    timeout_seconds = 30
    session_timeout = aiohttp.ClientTimeout(
        total=None, 
        sock_connect=timeout_seconds, 
        sock_read=timeout_seconds
    )

    async with aiohttp.ClientSession(timeout=session_timeout) as session:
        # Position nodes in the project
        nodex, nodey = -825, -375

        # Create nodes
        async with asyncio.TaskGroup() as tg1:
            node_tasks = []
            for switch in switches:
                task = tg1.create_task(
                    make_a_gns3_node(
                        switch, session, gns3_url, nodex, nodey, prj_id
                    )
                )
                node_tasks.append(task)
                
                # Update node positioning
                nodex += 150
                if nodex > 400:
                    nodex = -800
                    nodey += 200

        # Update switches with node information
        switches = [await task for task in node_tasks]

        # Configure Docker containers
        docker_client = aiodocker.Docker(url=f"http://{servername}:2375")
        try:
            async with asyncio.TaskGroup() as tg2:
                config_tasks = []
                for switch in switches:
                    task = tg2.create_task(
                        docker_api_config(switch, docker_client, servername)
                    )
                    config_tasks.append(task)
                
                # Wait for configuration tasks
                results = [await task for task in config_tasks]
                logger.info("Configuration copying completed for all switches")
        finally:
            await docker_client.close()

        # Establish connections between nodes
        async with asyncio.TaskGroup() as tg3:
            for connection in connections:
                a_node_id, b_node_id = _find_connection_nodes(switches, connection)
                
                if a_node_id and b_node_id:
                    try:
                        a_adapter = _parse_port(connection.port_a)
                        b_adapter = _parse_port(connection.port_b)

                        make_link_url = f"{gns3_url}projects/{prj_id}/links"
                        make_link_json = {
                            'nodes': [
                                {
                                    'adapter_number': int(a_adapter),
                                    'node_id': a_node_id, 
                                    'port_number': 0
                                },
                                {
                                    'adapter_number': int(b_adapter),
                                    'node_id': b_node_id, 
                                    'port_number': 0
                                }
                            ]
                        }

                        tg3.create_task(
                            gns3_post(
                                session, 
                                str(make_link_url), 
                                'post', 
                                jsondata=make_link_json
                            )
                        )
                    except ValueError as e:
                        logger.error(
                            f"Error parsing ports for connection "
                            f"{connection.switch_a}:{connection.port_a} -> "
                            f"{connection.switch_b}:{connection.port_b}: {e}"
                        )

        return "Virtual Network Lab is ready to run."

def _find_connection_nodes(
    switches: List[Switch], 
    connection: Connection
) -> Tuple[Optional[str], Optional[str]]:
    """
    Find node IDs for a given connection.

    Args:
        switches (List[Switch]): List of switches
        connection (Connection): Connection to find nodes for

    Returns:
        Tuple[Optional[str], Optional[str]]: Node IDs for switches A and B
    """
    a_node_id = next(
        (switch.gns3_node_id for switch in switches 
         if switch.lldp_system_name == connection.switch_a), 
        None
    )
    b_node_id = next(
        (switch.gns3_node_id for switch in switches 
         if switch.lldp_system_name == connection.switch_b), 
        None
    )
    return a_node_id, b_node_id

def _parse_port(port: str) -> str:
    """
    Parse port string to extract adapter number.

    Args:
        port (str): Port string (e.g., 'Ethernet1/1')

    Returns:
        str: Extracted adapter number

    Raises:
        ValueError: If port format is invalid
    """
    if not port.lower().startswith('ethernet'):
        raise ValueError(f"Invalid port format: {port}. Port must start with 'ethernet'")
    
    # Extract numbers after 'ethernet', before any '/'
    port_base = port.lower().split('/')[0]
    adapter_num = ''.join(filter(str.isdigit, port_base))
    
    if not adapter_num:
        raise ValueError(f"Could not extract adapter number from port: {port}")
    
    return adapter_num

async def docker_api_config(
    switch: Switch, 
    docker_client: aiodocker.Docker, 
    servername: str = 'localhost'
) -> str:
    """
    Configure Docker container with switch startup configuration.

    Args:
        switch (Switch): Switch object containing configuration
        docker_client (aiodocker.Docker): Docker client for API interactions
        servername (str, optional): Docker daemon hostname. Defaults to 'localhost'.

    Returns:
        str: Configuration status
    """
    try:
        logger.info(f"Starting configuration for switch {switch.name}")
        
        # Validate configuration
        if not switch.initial_config:
            logger.warning(f"Empty configuration for switch {switch.name}")
            return 'skipped - empty config'

        # Prepare configuration string
        config_string = '\n'.join(switch.initial_config)
        ascii_config = config_string.encode('ascii')
        
        # Create tar archive
        fh = BytesIO()
        with tarfile.open(fileobj=fh, mode='w') as tarch:
            info = tarfile.TarInfo('startup-config')
            info.size = len(ascii_config)
            bytes_to_go = BytesIO(ascii_config)
            tarch.addfile(info, bytes_to_go)
        
        fh.seek(0)
        archive_content = fh.read()
        
        logger.info(f"Configuration archive created for {switch.name}")

        # Async Docker API session
        async with aiohttp.ClientSession(base_url=f"http://{servername}:2375") as session:
            # Start container
            await _start_container(session, switch.docker_container_id)
            
            # Copy configuration to container
            await _copy_config_to_container(
                session, 
                switch.docker_container_id, 
                archive_content
            )
            
            # Move configuration file
            await _move_config_file(session, switch.docker_container_id)
            
            # Stop container
            await _stop_container(session, switch.docker_container_id)

        logger.info(f"Successfully configured switch {switch.name}")
        return 'success'

    except Exception as e:
        logger.error(f"Error configuring switch {switch.name}: {e}")
        return f'failed - {str(e)}'

async def _start_container(session: aiohttp.ClientSession, container_id: str) -> None:
    """
    Start a Docker container.

    Args:
        session (aiohttp.ClientSession): Async HTTP session
        container_id (str): Docker container ID

    Raises:
        ContainerConfigurationError: If container start fails
    """
    async with session.post(f"/containers/{container_id}/start") as response:
        if response.status not in [204, 304]:
            raise ContainerConfigurationError(
                f"Could not start container. Status: {response.status}"
            )
    
    # Wait for container to be ready
    start_time = asyncio.get_event_loop().time()
    while asyncio.get_event_loop().time() - start_time < 20:
        async with session.get(f"/containers/{container_id}/json") as response:
            if response.status == 200:
                container_info = await response.json()
                if container_info['State']['Running']:
                    return
        await asyncio.sleep(1)
    
    raise ContainerConfigurationError(f"Container {container_id} did not become ready")

async def _copy_config_to_container(
    session: aiohttp.ClientSession, 
    container_id: str, 
    archive_content: bytes
) -> None:
    """
    Copy configuration archive to container.

    Args:
        session (aiohttp.ClientSession): Async HTTP session
        container_id (str): Docker container ID
        archive_content (bytes): Configuration archive content

    Raises:
        ContainerConfigurationError: If file copy fails
    """
    headers = {'Content-Type': 'application/x-tar'}
    async with session.put(
        f"/containers/{container_id}/archive",
        params={'path': '/'},
        headers=headers,
        data=archive_content
    ) as response:
        if response.status != 200:
            raise ContainerConfigurationError(
                f"Error copying configuration. Status: {response.status}"
            )

async def _move_config_file(
    session: aiohttp.ClientSession, 
    container_id: str
) -> None:
    """
    Move configuration file to correct location in container.

    Args:
        session (aiohttp.ClientSession): Async HTTP session
        container_id (str): Docker container ID

    Raises:
        ContainerConfigurationError: If file move fails
    """
    # Create exec instance for mv command
    async with session.post(
        f"/containers/{container_id}/exec",
        json={
            "AttachStdout": True,
            "AttachStderr": True,
            "Cmd": ["mv", "/startup-config", "/mnt/flash/"]
        }
    ) as response:
        if response.status != 201:
            raise ContainerConfigurationError(
                f"Could not create mv exec. Status: {response.status}"
            )
        
        exec_data = await response.json()
        exec_id = exec_data['Id']

        # Start exec instance
        async with session.post(
            f"/exec/{exec_id}/start",
            json={"Detach": False, "Tty": False},
            headers={'Content-Type': 'application/json'}
        ) as exec_response:
            if exec_response.status != 200:
                raise ContainerConfigurationError(
                    f"mv command exec failed. Status: {exec_response.status}"
                )

async def _stop_container(
    session: aiohttp.ClientSession, 
    container_id: str
) -> None:
    """
    Stop a Docker container.

    Args:
        session (aiohttp.ClientSession): Async HTTP session
        container_id (str): Docker container ID

    Raises:
        ContainerConfigurationError: If container stop fails
    """
    async with session.post(f"/containers/{container_id}/stop") as response:
        if response.status not in [204, 304]:
            raise ContainerConfigurationError(
                f"Could not stop container. Status: {response.status}"
            )

async def make_a_gns3_node(
    switch: Switch, 
    session: aiohttp.ClientSession, 
    gns3_url: str, 
    nodex: int, 
    nodey: int, 
    prj_id: str
) -> Switch:
    """
    Create a GNS3 node for a switch.

    Args:
        switch (Switch): Switch to create a node for
        session (aiohttp.ClientSession): Async HTTP session
        gns3_url (str): Base URL for GNS3 API
        nodex (int): X coordinate for node placement
        nodey (int): Y coordinate for node placement
        prj_id (str): GNS3 project ID

    Returns:
        Switch: Updated switch with GNS3 node details
    """
    try:
        # Duplicate template
        async with session.post(
            f'{gns3_url}templates/{switch.gns3_template_id}/duplicate', 
            json=[]
        ) as response:
            response.raise_for_status()
            json_data = await response.json()
            tmp_template_id = json_data['template_id']

        # Update interface count
        async with session.put(
            f'{gns3_url}templates/{tmp_template_id}', 
            json={'adapters': switch.ethernet_interfaces + 1}
        ) as response:
            response.raise_for_status()

        # Create node
        async with session.post(
            f'{gns3_url}projects/{prj_id}/templates/{tmp_template_id}', 
            json={'x': nodex, 'y': nodey}
        ) as response:
            response.raise_for_status()
            json_data = await response.json()
            switch.gns3_node_id = json_data['node_id']

        # Delete temporary template
        async with session.delete(f'{gns3_url}templates/{tmp_template_id}'):
            pass

        # Rename node
        async with session.put(
            f'{gns3_url}projects/{prj_id}/nodes/{switch.gns3_node_id}', 
            json={'name': switch.name}
        ) as response:
            response.raise_for_status()

        # Get container ID
        async with session.get(
            f'{gns3_url}projects/{prj_id}/nodes/{switch.gns3_node_id}'
        ) as response:
            response.raise_for_status()
            json_data = await response.json()
            switch.docker_container_id = json_data['properties']['container_id']

        return switch

    except aiohttp.ClientResponseError as e:
        logger.error(f"GNS3 node creation failed: {e}")
        raise GNS3WorkerError(f"Failed to create GNS3 node: {e}")

async def gns3_post(
    session: aiohttp.ClientSession, 
    url: str, 
    method: str, 
    **kwargs
) -> None:
    """
    Send an async request to GNS3 server.

    Args:
        session (aiohttp.ClientSession): Async HTTP session
        url (str): URL to send request to
        method (str): HTTP method (get, post, put)
        **kwargs: Additional arguments for the request

    Raises:
        GNS3WorkerError: If request fails
    """
    try:
        jsondata = kwargs.get('jsondata', {})
        
        if method == 'post':
            async with session.post(url, json=jsondata) as response:
                response.raise_for_status()
        elif method == 'get':
            async with session.get(url, json=jsondata) as response:
                response.raise_for_status()
        elif method == 'put':
            async with session.put(url, json=jsondata) as response:
                response.raise_for_status()
        
        # Small delay to prevent overwhelming the server
        await asyncio.sleep(0.2)

    except aiohttp.ClientResponseError as e:
        logger.error(f"GNS3 API request failed: {e}")
        raise GNS3WorkerError(f"API request failed: {e}")



'''This is a wrapper script for vnlOnDamend to allow it be be invoked from outside of the
 Python interpreter.  To use it, copy the contents of the ptovnetlab folder to the host you
will run it from.  Then move the vnlOnDemandCli.py file *up* one level, so that is is in the
 same directory as the vnlOnDemand folder.  Refer to the docs for additional details.'''


import ptovnetlab.ptovnetlab as ptovnetlab
import sys

if __name__ == '__main__':
    kwdict = {}
    for arg in sys.argv[1:]:
        splarg = arg.split('=')
        if splarg[0] == 'switchlist':
            splargalt = []
            for swname in splarg[1].split():
                splargalt.append(swname)
            kwdict[splarg[0]] = splargalt
        else:
            kwdict[splarg[0]] = splarg[1]
    kwdict['runtype'] = 'script'
    ptovnetlab.p_to_v(**kwdict)



#!/usr/bin/env python3
"""
EOS to cEOS Configuration Sanitization Module

This module provides functionality for converting Arista hardware appliance 
EOS configurations to be suitable for use in a cEOS (containerized EOS) 
lab environment.
"""

import logging
from typing import List, Optional

from ptovnetlab.data_classes import Switch

# Configure logging
logging.basicConfig(
    level=logging.INFO, 
    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s'
)
logger = logging.getLogger(__name__)

class EosSanitizerError(Exception):
    """Custom exception for EOS to cEOS configuration sanitization errors."""
    pass

def eos_to_ceos(switch: Switch) -> Switch:
    """
    Convert an Arista switch configuration for cEOS lab environment.

    This function modifies the configuration to be compatible with 
    containerized EOS, including:
    - Replacing management interface names
    - Removing incompatible configuration lines
    - Handling interface configurations
    - Applying system MAC address

    Args:
        switch (Switch): Switch object containing the original configuration

    Returns:
        Switch: Updated Switch object with cEOS-compatible configuration
    """
    try:
        # Validate input
        if not switch or not switch.initial_config:
            raise EosSanitizerError("Invalid switch configuration: Empty config")

        # Configuration lines to be commented out in cEOS lab environment
        bad_starts = [
            'radius', 'username', 'aaa', 'ip radius', 'hardware speed', 
            'queue', 'server ', 'ntp server', 'daemon TerminAttr', 
            '   exec /usr/bin/TerminAttr'
        ]

        # Count Ethernet interfaces
        switch.ethernet_interfaces = _count_ether_interfaces(switch.initial_config)

        # Process configuration lines
        sanitized_config = _sanitize_config_lines(
            switch.initial_config, 
            bad_starts, 
            switch.system_mac
        )

        # Update switch with sanitized configuration
        switch.initial_config = sanitized_config

        logger.info(f"Converted configuration for switch: {switch.name}")
        return switch

    except Exception as e:
        logger.error(f"Configuration conversion failed for {switch.name}: {e}")
        raise EosSanitizerError(f"Conversion error: {e}")

def _sanitize_config_lines(
    config_lines: List[str], 
    bad_starts: List[str], 
    system_mac: str
) -> List[str]:
    """
    Sanitize configuration lines for cEOS compatibility.

    Args:
        config_lines (List[str]): Original configuration lines
        bad_starts (List[str]): Prefixes of lines to be commented out
        system_mac (str): System MAC address to be applied

    Returns:
        List[str]: Sanitized configuration lines
    """
    sanitized_lines = []
    mgt_port_str = 'Management0'

    for line in config_lines:
        # Replace Management1/Management0 interface names
        line = line.replace('Management1', mgt_port_str)
        line = line.replace('Management0', mgt_port_str)

        # Comment out lines starting with bad prefixes
        if any(line.startswith(bad) for bad in bad_starts):
            line = f"!removed_for_cEOS-lab| {line}"

        # Handle spurious interface configurations
        if line.startswith('interface Ethernet'):
            # Remove breakout interfaces (/2, /3, /4)
            if any(f'/{n}' in line for n in [2, 3, 4]):
                line = f"!{line}"
            # Simplify interface names (remove subinterface)
            else:
                line = line.split('/')[0]

        sanitized_lines.append(line)

    # Apply system MAC address configuration
    sanitized_lines = _apply_sys_mac(sanitized_lines, system_mac)

    return sanitized_lines

def _count_ether_interfaces(config: List[str]) -> int:
    """
    Count the number of Ethernet interfaces in the configuration.

    Args:
        config (List[str]): List of configuration lines

    Returns:
        int: Number of Ethernet interfaces
    """
    return sum(
        1 for line in config 
        if (line.startswith('interface Ethernet') and 
            not any(line.endswith(f'/{n}') for n in [2, 3, 4]))
    )

def _apply_sys_mac(config: List[str], sys_mac: str) -> List[str]:
    """
    Append system MAC address configuration to the config.

    Args:
        config (List[str]): Original configuration lines
        sys_mac (str): System MAC address to apply

    Returns:
        List[str]: Updated configuration with system MAC handler
    """
    # Create event handler to set system MAC address
    sys_mac_snippet = [
        '',
        'event-handler onStartup',
        ' trigger on-boot',
        ' action bash',
        f'      var_sysmac=\'{sys_mac}\'',
        '  echo $var_sysmac > /mnt/flash/system_mac_address',
        '  truncate -s -1 /mnt/flash/system_mac_address',
        '  EOF'
    ]

    # Remove the last line ('end') and append system MAC config
    if config and config[-1] == 'end':
        config.pop()
        config.extend(sys_mac_snippet)
        config.append('end')

    return config



#!/usr/bin/env python3
"""
Arista Network Device Polling Module

This module provides asynchronous functionality for polling multiple Arista 
network switches concurrently using pyeapi, extracting device information 
and LLDP connection details.
"""

import asyncio
import logging
from concurrent.futures import ThreadPoolExecutor
from typing import List, Tuple, Optional

import pyeapi
from ptovnetlab.data_classes import Switch, Connection

# Configure logging
logging.basicConfig(
    level=logging.INFO, 
    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s'
)
logger = logging.getLogger(__name__)

class AristaPollerError(Exception):
    """Custom exception for Arista Poller related errors."""
    pass

def validate_switch_credentials(switch: str, username: str, password: str) -> bool:
    """
    Validate switch connection credentials.

    Args:
        switch (str): Switch hostname or IP address
        username (str): Authentication username
        password (str): Authentication password

    Returns:
        bool: True if credentials are valid, False otherwise
    """
    try:
        pyeapi.client.config.clear()
        pyeapi.client.config.add_connection(
            switch, 
            host=switch, 
            transport='https',
            username=username, 
            password=password
        )
        node = pyeapi.connect_to(switch)
        # Attempt a simple command to verify connection
        node.enable("show version")
        return True
    except Exception as e:
        logger.error(f"Credential validation failed for {switch}: {e}")
        return False

def invoker(
    switchlist_in: List[str], 
    uname_in: str, 
    passwd_in: str,
    runtype_in: str
) -> Tuple[List[Switch], List[Connection]]:
    """
    Synchronous entry point for switch polling.

    Args:
        switchlist_in (List[str]): List of Arista switches to interrogate
        uname_in (str): Username for switch authentication
        passwd_in (str): Password for switch authentication
        runtype_in (str): Type of polling run (e.g., 'discovery', 'update')

    Returns:
        Tuple[List[Switch], List[Connection]]: Polled switches and their connections
    """
    try:
        # Validate credentials for all switches before polling
        valid_switches = [
            switch for switch in switchlist_in 
            if validate_switch_credentials(switch, uname_in, passwd_in)
        ]

        if not valid_switches:
            raise AristaPollerError("No valid switches found for polling")

        switches, connections = asyncio.run(
            main(valid_switches, uname_in, passwd_in, runtype_in)
        )
        return switches, connections
    except Exception as e:
        logger.error(f"Switch polling failed: {e}")
        raise

async def main(
    switchlist_in2: List[str], 
    uname_in2: str, 
    passwd_in2: str,
    runtype_in2: str
) -> Tuple[List[Switch], List[Connection]]:
    """
    Asynchronously poll multiple Arista switches.

    Args:
        switchlist_in2 (List[str]): Switches to poll
        uname_in2 (str): Authentication username
        passwd_in2 (str): Authentication password
        runtype_in2 (str): Type of polling run

    Returns:
        Tuple[List[Switch], List[Connection]]: Polled switches and their connections
    """
    # Set the maximum number of worker threads
    loop = asyncio.get_running_loop()
    loop.set_default_executor(ThreadPoolExecutor(max_workers=20))
    
    logger.info(f'Polling {len(switchlist_in2)} Arista switches via EOS API...')
    
    # Create tasks for each switch
    tasks = [
        asyncio.create_task(
            asyncio.to_thread(
                get_sw_data, 
                switch, 
                uname_in2, 
                passwd_in2, 
                sw_index
            )
        ) 
        for sw_index, switch in enumerate(switchlist_in2)
    ]

    # Gather the data from all EAPI polling threads
    try:
        answers = await asyncio.gather(*tasks)
        logger.info('Finished polling switches.')

        switches = []
        connections = []
        
        # Process the gathered data
        for val in answers:
            switch, lldp_connections = val
            switches.append(switch)
            connections.extend(lldp_connections)

        return switches, connections

    except Exception as e:
        logger.error(f"Error during switch polling: {e}")
        raise AristaPollerError(f"Switch polling failed: {e}")

def get_sw_data(
    switch3: str, 
    uname_in3: str, 
    passwd_in3: str, 
    sw_cntr3_in: int
) -> Tuple[Switch, List[Connection]]:
    """
    Retrieve switch data and LLDP connections for a single switch.

    Args:
        switch3 (str): Switch hostname or IP to interrogate
        uname_in3 (str): Authentication username
        passwd_in3 (str): Authentication password
        sw_cntr3_in (int): Switch index for logging purposes

    Returns:
        Tuple[Switch, List[Connection]]: Switch object and its LLDP connections
    """
    try:
        # Clear any existing pyeapi.client.config
        pyeapi.client.config.clear()
        
        # Build the pyeapi.client.config object
        pyeapi.client.config.add_connection(
            switch3, 
            host=switch3, 
            transport='https',
            username=uname_in3, 
            password=passwd_in3
        )
        
        # Connect to the switch
        node = pyeapi.connect_to(switch3)
        
        # Get JSON-formatted results of several 'show...' commands
        eos_output = node.enable(
            ("show version", "show lldp neighbors", "show lldp local-info"), 
            format="json"
        )
        
        # Create Switch object
        switch = Switch(
            name=switch3,
            model=eos_output[0]["result"]["modelName"],
            eos_version=eos_output[0]["result"]["version"],
            system_mac=eos_output[0]["result"]["systemMacAddress"],
            serial_number=eos_output[0]["result"]["serialNumber"],
            lldp_system_name=eos_output[2]["result"]["systemName"],
            ethernet_interfaces=0,  # Will be set by arista_sanitizer
            gns3_template_id='',   # Will be set by gns3_worker
            gns3_node_id='',       # Will be set by gns3_worker
            docker_container_id='', # Will be set by gns3_worker
            initial_config=node.running_config.splitlines()
        )

        # Create Connection objects from LLDP neighbors
        connections = [
            Connection(
                switch_a=str(eos_output[2]["result"]["systemName"]),
                port_a=str(value["port"]),
                switch_b=str(value["neighborDevice"]),
                port_b=str(value["neighborPort"])
            )
            for value in eos_output[1]["result"]["lldpNeighbors"]
        ]

        logger.info(f"Finished polling switch: {switch3}")
        return switch, connections

    except Exception as e:
        logger.error(f"Error polling switch {switch3}: {e}")
        raise AristaPollerError(f"Failed to poll switch {switch3}: {e}")



"""PTovNetLab: Automated Network Lab Creation

This module converts physical network switch configurations into a GNS3 virtual lab.
"""

import sys
import logging
from typing import List, Dict, Optional, Tuple
from getpass import getpass
import requests

from ptovnetlab import arista_poller, arista_sanitizer, gns3_worker
from ptovnetlab.data_classes import Switch, Connection

# Custom Exceptions
class PTovNetLabError(Exception):
    """Base exception for PTovNetLab errors."""
    pass

class InputValidationError(PTovNetLabError):
    """Raised when input validation fails."""
    pass

class NetworkConfigurationError(PTovNetLabError):
    """Raised when network configuration processing fails."""
    pass

# Configure logging
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s'
)
logger = logging.getLogger(__name__)

def validate_input(
    filename: Optional[str] = None, 
    switchlist: Optional[List[str]] = None, 
    prj_name: Optional[str] = None, 
    servername: Optional[str] = None
) -> None:
    """
    Validate input parameters for the virtual lab creation.

    Args:
        filename: Optional path to a file containing switch names
        switchlist: Optional list of switch names
        prj_name: Optional project name
        servername: Optional GNS3 server name

    Raises:
        InputValidationError: If input parameters are invalid
    """
    if filename and switchlist:
        raise InputValidationError(
            "Cannot provide both filename and switchlist. Choose one method."
        )
    
    if not (filename or switchlist):
        logger.warning("No switch sources provided. Entering interactive mode.")

    if not prj_name:
        raise InputValidationError("Project name is required")

    if not servername:
        raise InputValidationError("GNS3 server name is required")

def collect_switch_list(
    filename: Optional[str] = None, 
    switchlist: Optional[List[str]] = None
) -> List[str]:
    """
    Collect switch list from file or interactive input.

    Args:
        filename: Optional path to a file containing switch names
        switchlist: Optional list of switch names

    Returns:
        List of switch names
    """
    if filename:
        try:
            with open(filename, 'r') as f:
                return [line.strip() for line in f if line.strip()]
        except IOError as e:
            raise InputValidationError(f"Error reading switch list file: {e}")
    
    if switchlist:
        return switchlist

    # Interactive input
    logger.info("Enter switch names (press Enter without input to finish):")
    interactive_list = []
    while True:
        switch = input().strip()
        if not switch:
            break
        interactive_list.append(switch)
    
    return interactive_list

def authenticate_switches() -> Tuple[str, str]:
    """
    Authenticate switches interactively.

    Returns:
        Tuple of (username, password)
    """
    username = input('Enter username for Arista EOS login: ')
    passwd = getpass('Enter password for Arista EOS login: ')
    return username, passwd

def process_switch_configurations(
    switches: List[Switch], 
    connections: List[Connection]
) -> Tuple[List[Switch], List[Connection]]:
    """
    Process and sanitize switch configurations.

    Args:
        switches: List of switches
        connections: List of connections

    Returns:
        Processed switches and connections
    """
    # Sanitize configurations
    for switch in switches:
        switch = arista_sanitizer.eos_to_ceos(switch)

    # Filter connections involving only the current switches
    our_lldp_ids = {switch.lldp_system_name for switch in switches}
    connections = [
        conn for conn in connections 
        if conn.switch_a in our_lldp_ids and conn.switch_b in our_lldp_ids
    ]

    # Remove duplicate connections
    unique_connections = []
    for conn in connections:
        if not any(
            c.switch_a == conn.switch_b and 
            c.switch_b == conn.switch_a and
            c.port_a == conn.port_b and 
            c.port_b == conn.port_a 
            for c in unique_connections
        ):
            unique_connections.append(conn)

    # Clean management interfaces
    for conn in unique_connections:
        conn.port_a = 'ethernet0' if conn.port_a.lower().startswith('management') else conn.port_a
        conn.port_b = 'ethernet0' if conn.port_b.lower().startswith('management') else conn.port_b

    return switches, unique_connections

def p_to_v(**kwargs) -> str:
    """
    Convert physical network to virtual lab.

    Args:
        **kwargs: Flexible keyword arguments for configuration

    Returns:
        URL to the created GNS3 project
    """
    try:
        # Extract and validate inputs
        filename = kwargs.get('filename', '')
        switchlist = kwargs.get('switchlist', [])
        username = kwargs.get('username', '')
        passwd = kwargs.get('passwd', '')
        servername = kwargs.get('servername', '')
        prj_name = kwargs.get('prjname', '')
        run_type = kwargs.get('runtype', 'module')

        validate_input(filename, switchlist, prj_name, servername)
        
        # Collect switch list
        switchlist = collect_switch_list(filename, switchlist)
        
        # Authenticate if credentials not provided
        if not (username and passwd):
            username, passwd = authenticate_switches()

        # Poll switch configurations
        switches, connections = arista_poller.invoker(switchlist, username, passwd, run_type)
        
        # Process switch configurations
        switches, connections = process_switch_configurations(switches, connections)

        # Set GNS3 URLs
        gns3_url = f'http://{servername}:3080/v2/'
        gns3_url_noapi = f'http://{servername}:3080/static/web-ui/server/1/project/'

        # Get and map GNS3 templates
        r = requests.get(gns3_url + 'templates', auth=('admin', 'admin'), timeout=20)
        image_map = {
            x['image'].lower(): x['template_id'] 
            for x in r.json() 
            if x['template_type'] == 'docker'
        }

        # Set template IDs for switches
        for switch in switches:
            eos_version = 'ceos:' + switch.eos_version.lower().split('-')[0]
            if eos_version in image_map:
                switch.gns3_template_id = image_map[eos_version]

        # Create GNS3 project
        gnsprj_id = requests.post(
            gns3_url + 'projects', 
            json={'name': prj_name},
            timeout=20
        ).json()['project_id']

        # Create nodes and connections
        gns3_worker.invoker(servername, gns3_url, switches, gnsprj_id, connections)

        logger.info(f"Successfully created GNS3 project: {prj_name}")
        return gns3_url_noapi + gnsprj_id

    except (InputValidationError, NetworkConfigurationError) as e:
        logger.error(f"Configuration error: {e}")
        raise
    except Exception as e:
        logger.error(f"Unexpected error in virtual lab creation: {e}")
        raise PTovNetLabError(f"Virtual lab creation failed: {e}")

def main():
    """
    Main entry point for script execution.
    """
    try:
        kwdict = {}
        for arg in sys.argv[1:]:
            splarg = arg.split('=')
            if splarg[0] == 'switchlist':
                kwdict[splarg[0]] = splarg[1].split()
            else:
                kwdict[splarg[0]] = splarg[1]
        kwdict['runtype'] = 'script'
        p_to_v(**kwdict)
    except PTovNetLabError as e:
        logger.error(f"PTovNetLab execution failed: {e}")
        sys.exit(1)

if __name__ == '__main__':
    main()



from dataclasses import dataclass, field
from typing import List, Optional

@dataclass
class Switch:
    """Represents a switch being modeled in the virtual lab."""
    name: str  # Switch name received as input argument
    model: str  # Switch model
    eos_version: str  # EOS version
    system_mac: str  # System MAC address
    serial_number: str  # Serial number
    lldp_system_name: str  # LLDP system name
    ethernet_interfaces: int  # Number of Ethernet interfaces
    gns3_template_id: str  # GNS3 image-template ID
    gns3_node_id: str  # GNS3 node-ID
    docker_container_id: str  # Docker container ID
    initial_config: List = field(default_factory=list)  # The modeled device's configuration
    vendor_platform: str = ''  # Switch vendor/platform (optional)
    qemu_vm_id: str = ''  # QEMU VM ID (optional)

@dataclass
class Connection:
    """Represents a connection between two switches in the virtual lab."""
    switch_a: str  # LLDP system name of the first switch
    port_a: str  # Port on the first switch
    switch_b: str  # LLDP system name of the second switch
    port_b: str  # Port on the second switch
